"""This 'experimental' package is a place for code to live that might change significantly or might
not be fully thought out or feature rich enough to include in the firmware proper. It exists for the
purpose of allowing developers to experiment, colaborate, write scripts against it and provide
feedback. Code here should slowly evolve and stabilize.

Once code here has proven to be both stable and useful we can promote it out of the experimental
into the firmware library package and update calling scripts.

Code that is neither stable nor useful should be removed.
"""
